import numpy as np
import rdflib

path = '/Users/mehdi/PycharmProjects/PyKEEN/data/corpora/rdf.nt'
g = rdflib.Graph()
g.parse(path, format="nt")


for s, p, o in g:
    print(str(s))
    print(type(p))
    print(o)
    print('-----------')
    exit(0)




import torch
import torch.nn.functional as F
import torch.nn as nn
import numpy as np
import torch.optim as optim

class TransD(nn.Module):

    def __init__(self):
        super(TransD, self).__init__()
        self.num_entities = 2
        self.num_relations = 2
        self.entity_embedding_dim = 2
        self.relation_embedding_dim = 3
        self.margin_loss = 1
  

        # A simple lookup table that stores embeddings of a fixed dictionary and size
        self.entity_embeddings = nn.Embedding(self.num_entities, self.entity_embedding_dim, max_norm=1)
        self.relation_embeddings = nn.Embedding(self.num_relations, self.relation_embedding_dim, max_norm=1)
        self.entity_projections = nn.Embedding(self.num_entities, self.entity_embedding_dim)
        self.relation_projections = nn.Embedding(self.num_relations, self.relation_embedding_dim)

        self.criterion = nn.MarginRankingLoss(margin=self.margin_loss, size_average=True)
        self.scoring_fct_norm = 1
        # self._initialize()

    def _compute_scores(self, h_embs, r_embs, t_embs):
        """

        :param h_embs:
        :param r_embs:
        :param t_embs:
        :return:
        """

        # Add the vector element wise
        sum_res = h_embs + r_embs - t_embs
        distances = torch.norm(sum_res, dim=1, p=self.scoring_fct_norm).view(size=(-1,))
        distances = torch.mul(distances, distances)

        return distances

    def _compute_loss(self, pos_scores, neg_scores):
        """

        :param pos_scores:
        :param neg_scores:
        :return:
        """

        # y == -1 indicates that second input to criterion should get a larger loss
        # y = torch.Tensor([-1]).cuda()
        # NOTE: y = 1 is important
        # y = torch.tensor([-1], dtype=torch.float)
        y = np.repeat([-1], repeats=pos_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float)

        # Scores for the psotive and negative triples
        pos_scores = torch.tensor(pos_scores, dtype=torch.float)
        neg_scores = torch.tensor(neg_scores, dtype=torch.float)

        loss = self.criterion(pos_scores, neg_scores, y)

        return loss

    def _project_entities(self, entity_embs, entity_proj_vecs, relation_projs):
        # batch_size = entity_embs.shape[0]
        # identity_matrices = torch.eye(batch_size,self.relation_embedding_dim,self.entity_embedding_dim)
        entity_embs = entity_embs
        relation_projs = relation_projs.unsqueeze(-1)
        entity_proj_vecs = entity_proj_vecs.unsqueeze(-1).permute([0,2,1])

        transfer_matrices = torch.matmul(relation_projs,entity_proj_vecs)



        # transfer_matrices = torch.einsum('nm1,nk1->nmk',
        #                                  [relation_projs, entity_proj_vecs])  # TODO: Check + identity_matrices


        # a = relation_projs.view(-1,self.relation_embedding_dim)
        # torch.matmul(transfer_matrices)

        projected_entity_embs = torch.einsum('nmk,nk->nm', [transfer_matrices, entity_embs])
        # projected_entity_embs = F.normalize(projected_entity_embs, 2, 1)

        return projected_entity_embs

    # def _initialize(self):
    #     lower_bound = -6 / np.sqrt(self.entity_embedding_dim)
    #     upper_bound = 6 / np.sqrt(self.entity_embedding_dim)
    #     nn.init.uniform_(self.entity_embeddings.weight.data, a=lower_bound, b=upper_bound)
    #     nn.init.uniform_(self.relation_embeddings.weight.data, a=lower_bound, b=upper_bound)
    #
    #     norms = torch.norm(self.relation_embeddings.weight, p=2, dim=1).data
    #     self.relation_embeddings.weight.data = self.relation_embeddings.weight.data.div(
    #         norms.view(self.num_relations, 1).expand_as(self.relation_embeddings.weight))

    def forward(self, batch_positives, batch_negatives):
        pos_heads = batch_positives[:, 0:1]
        pos_relations = batch_positives[:, 1:2]
        pos_tails = batch_positives[:, 2:3]

        neg_heads = batch_negatives[:, 0:1]
        neg_relations = batch_negatives[:, 1:2]
        neg_tails = batch_negatives[:, 2:3]

        pos_h_embs = self.entity_embeddings(pos_heads).view(-1, self.entity_embedding_dim)
        pos_r_embs = self.relation_embeddings(pos_relations).view(-1, self.relation_embedding_dim)
        pos_t_embs = self.entity_embeddings(pos_tails).view(-1, self.entity_embedding_dim)

        pos_h_proj_vec_embs = self.entity_projections(pos_heads).view(-1, self.entity_embedding_dim)
        pos_r_projs_embs = self.relation_projections(pos_relations).view(-1, self.relation_embedding_dim)
        pos_t_proj_vec_embs = self.entity_projections(pos_tails).view(-1, self.entity_embedding_dim)

        neg_h_embs = self.entity_embeddings(neg_heads).view(-1, self.entity_embedding_dim)
        neg_r_embs = self.relation_embeddings(neg_relations).view(-1, self.relation_embedding_dim)
        neg_t_embs = self.entity_embeddings(neg_tails).view(-1, self.entity_embedding_dim)

        neg_h_proj_vec_embs = self.entity_projections(neg_heads).view(-1, self.entity_embedding_dim)
        neg_r_projs_embs = self.relation_projections(neg_relations).view(-1, self.relation_embedding_dim)
        neg_t_proj_vec_embs = self.entity_projections(neg_tails).view(-1, self.entity_embedding_dim)

        # Project entities
        proj_pos_heads = self._project_entities(pos_h_embs, pos_h_proj_vec_embs, pos_r_projs_embs)
        proj_pos_tails = self._project_entities(pos_t_embs, pos_t_proj_vec_embs, pos_r_projs_embs)

        proj_neg_heads = self._project_entities(neg_h_embs, neg_h_proj_vec_embs, neg_r_projs_embs)
        proj_neg_tails = self._project_entities(neg_t_embs, neg_t_proj_vec_embs, neg_r_projs_embs)

        pos_scores = self._compute_scores(h_embs=proj_pos_heads, r_embs=pos_r_embs, t_embs=proj_pos_tails)
        neg_scores = self._compute_scores(h_embs=proj_neg_heads, r_embs=neg_r_embs, t_embs=proj_neg_tails)

        # pos_scores = self._compute_scores(h_embs=pos_h_embs, r_embs=pos_r_embs, t_embs=pos_t_embs)
        # neg_scores = self._compute_scores(h_embs=neg_h_embs, r_embs=neg_r_embs, t_embs=neg_t_embs)


        loss = self._compute_loss(pos_scores=pos_scores, neg_scores=neg_scores)

        return loss
    
if __name__ == '__main__':
    pos_heads = torch.tensor([0, 1])
    pos_rels = torch.tensor([0, 0])
    pos_tails = torch.tensor([0, 0])


    # criterion = nn.MarginRankingLoss(margin=1, size_average=True)

    num_entities = 2
    num_relations = 2
    relation_embedding_dim = 3
    entity_embedding_dim = 2
    batch_size = 2

    pos_heads = torch.tensor([0, 1])
    pos_rels = torch.tensor([0, 0])
    pos_tails = torch.tensor([0, 0])

    neg_heads = torch.tensor([1, 0])
    neg_rels = torch.tensor([1, 1])
    neg_tails = torch.tensor([1, 0])

    pos_batch = torch.tensor([[0,0,0],[1,0,0]],dtype=torch.long)
    neg_batch = torch.tensor([[1,1,1], [0,1,0]], dtype=torch.long)

    entity_embeddings = nn.Embedding(num_entities, entity_embedding_dim,max_norm=1)
    relation_embeddings = nn.Embedding(num_relations, relation_embedding_dim,max_norm=1)
    entity_projections = nn.Embedding(num_entities, entity_embedding_dim)
    relation_projections = nn.Embedding(num_relations, relation_embedding_dim)
    model = TransD()
    optimizer = optim.SGD(model.parameters(), lr=0.1)
    for i in range(5):
        loss = model(pos_batch,neg_batch)
        loss.backward()
        print(loss)

        optimizer.step()
        optimizer.zero_grad()
        
        # pos_h_embs = entity_embeddings(pos_heads).view(-1, entity_embedding_dim)
        # pos_r_embs = relation_embeddings(pos_rels).view(-1, relation_embedding_dim)
        # pos_t_embs = entity_embeddings(pos_tails).view(-1, entity_embedding_dim)
        #
        # pos_h_projs_embs = entity_projections(pos_heads).view(-1, entity_embedding_dim)
        # pos_r_projs_embs = relation_projections(pos_rels).view(-1, relation_embedding_dim)
        # pos_t_projs_embs = entity_projections(pos_tails).view(-1, entity_embedding_dim)
        #
        # neg_h_embs = entity_embeddings(pos_heads).view(-1, entity_embedding_dim)
        # neg_r_embs = relation_embeddings(pos_rels).view(-1, relation_embedding_dim)
        # neg_t_embs = entity_embeddings(pos_tails).view(-1, entity_embedding_dim)
        #
        # neg_h_projs_embs = entity_projections(neg_heads).view(-1, entity_embedding_dim)
        # neg_r_projs_embs = relation_projections(neg_rels).view(-1, relation_embedding_dim)
        # neg_t_projs_embs = entity_projections(neg_tails).view(-1, entity_embedding_dim)
        #
        # pos_transfer_matrices = torch.einsum('nm,nk->nmk', [pos_r_projs_embs, pos_h_projs_embs])
        # pos_projected_entity_embs = torch.einsum('nmk,nk->nm', [pos_transfer_matrices, pos_h_embs])
        #
        # pos_sum_res = pos_projected_entity_embs + pos_r_embs - pos_projected_entity_embs
        # pos_distances = torch.norm(pos_sum_res, dim=1, p=1).view(size=(-1,))
        # pos_distances = torch.mul(pos_distances, pos_distances)
        #
        # neg_transfer_matrices = torch.einsum('nm,nk->nmk', [neg_r_projs_embs, neg_h_projs_embs])
        # neg_projected_entity_embs = torch.einsum('nmk,nk->nm', [neg_transfer_matrices, neg_h_embs])
        #
        # neg_sum_res = neg_projected_entity_embs + neg_r_embs - neg_projected_entity_embs
        # neg_distances = torch.norm(neg_sum_res, dim=1, p=1).view(size=(-1,))
        # neg_distances = torch.mul(neg_distances, neg_distances)
        #
        #
        # y = np.repeat([-1], repeats=2)
        # y = torch.tensor(y, dtype=torch.float)
        #
        # # Scores for the psotive and negative triples
        # pos_scores = torch.tensor(pos_distances, dtype=torch.float)
        #
        # loss = criterion(pos_scores, neg_distances, y)
        #
        # print(loss)
        #
        # loss.backward()
        #






import torch
import torch.autograd
import torch.nn as nn

if __name__ == '__main__':
    mlp = nn.Linear(3,5)
    embedding_dim = 3

    a = torch.tensor([[1,2,3],[2,4,6]],dtype=torch.float)

    print(mlp(a).shape)

    b = torch.cat([a,a],dim=1)

    print(b)

    mlp = nn.Sequential(
        nn.Linear(2 * embedding_dim, embedding_dim),
        nn.ReLU(),
        nn.Dropout(p=0.2),
        nn.Linear(embedding_dim, 1),
    )

    print()
    print(mlp(b).shape)
    print(mlp(b))


# -*- coding: utf-8 -*-

"""Implementation of the basic pipeline."""

import logging
from dataclasses import dataclass
from typing import Dict, Mapping

import numpy as np
import torch
from sklearn.model_selection import train_test_split
from torch.nn import Module

from pykeen.constants import *
from pykeen.hyper_parameter_optimizer.random_search_optimizer import RandomSearchHPO
from pykeen.utilities.evaluation_utils.metrics_computations import compute_metric_results
from pykeen.utilities.initialization_utils.module_initialization_utils import get_kg_embedding_model
from pykeen.utilities.train_utils import train_model
from pykeen.utilities.triples_creation_utils.instance_creation_utils import create_mapped_triples, create_mappings

__all__ = ['Pipeline']

log = logging.getLogger(__name__)


@dataclass
class PipelineResult:  # TODO replace in Pipeline.run()
    """Results from the pipeline."""


class Pipeline(object):
    def __init__(self, config: Dict):
        self.config: Dict = config
        self.seed: int = config[SEED]
        self.entity_to_id = None
        self.rel_to_id = None
        self.device_name = (
            'cuda:0'
            if torch.cuda.is_available() and self.config[PREFERRED_DEVICE] == GPU else
            CPU
        )
        self.device = torch.device(self.device_name)


    @staticmethod
    def _use_hpo(config):
        return config[EXECUTION_MODE] == HPO_MODE

    @property
    def is_evaluation_required(self) -> bool:
        return TEST_SET_PATH in self.config or TEST_SET_RATIO in self.config

    def run(self) -> Mapping:
        """Run this pipeline."""
        if self._use_hpo(self.config):  # Hyper-parameter optimization mode
            mapped_pos_train_triples, mapped_pos_test_triples = self._get_train_and_test_triples()

            (trained_model,
             loss_per_epoch,
             entity_to_embedding,
             relation_to_embedding,
             eval_summary,
             params) = RandomSearchHPO.run(
                mapped_train_triples=mapped_pos_train_triples,
                mapped_test_triples=mapped_pos_test_triples,
                entity_to_id=self.entity_to_id,
                rel_to_id=self.rel_to_id,
                config=self.config,
                device=self.device,
                seed=self.seed,
            )
        else:  # Training Mode
            if self.is_evaluation_required:
                mapped_pos_train_triples, mapped_pos_test_triples = self._get_train_and_test_triples()
            else:
                mapped_pos_train_triples = self._get_train_triples()

            all_entities = np.array(list(self.entity_to_id.values()))

            # Initialize KG embedding model
            self.config[NUM_ENTITIES] = len(self.entity_to_id)
            self.config[NUM_RELATIONS] = len(self.rel_to_id)
            self.config[PREFERRED_DEVICE] = CPU if self.device_name == CPU else GPU
            kg_embedding_model: Module = get_kg_embedding_model(config=self.config)

            batch_size = self.config[BATCH_SIZE]
            num_epochs = self.config[NUM_EPOCHS]
            learning_rate = self.config[LEARNING_RATE]

            log.info("-------------Train KG Embeddings-------------")
            trained_model, loss_per_epoch = train_model(
                kg_embedding_model=kg_embedding_model,
                all_entities=all_entities,
                learning_rate=learning_rate,
                num_epochs=num_epochs,
                batch_size=batch_size,
                pos_triples=mapped_pos_train_triples,
                device=self.device,
                seed=self.seed,
            )

            params = self.config
            eval_summary = None

            if self.is_evaluation_required:
                log.info("-------------Start Evaluation-------------")

                mean_rank, hits_at_k = compute_metric_results(
                    all_entities=all_entities,
                    kg_embedding_model=kg_embedding_model,
                    mapped_train_triples=mapped_pos_train_triples,
                    mapped_test_triples=mapped_pos_test_triples,
                    device=self.device,
                    filter_neg_triples=self.config[FILTER_NEG_TRIPLES],
                )

                eval_summary = OrderedDict()
                eval_summary[MEAN_RANK] = mean_rank
                eval_summary[HITS_AT_K] = hits_at_k

        # Prepare Output
        id_to_entity = {value: key for key, value in self.entity_to_id.items()}
        id_to_rel = {value: key for key, value in self.rel_to_id.items()}
        entity_to_embedding = {
            id_to_entity[id]: embedding.detach().cpu().numpy()
            for id, embedding in enumerate(trained_model.entity_embeddings.weight)
        }

        if self.config[KG_EMBEDDING_MODEL_NAME] in (SE_NAME, UM_NAME):
            relation_to_embedding = None
        else:
            relation_to_embedding = {
                id_to_rel[id]: embedding.detach().cpu().numpy()
                for id, embedding in enumerate(trained_model.relation_embeddings.weight)
            }

        results = OrderedDict()
        results[TRAINED_MODEL] = trained_model
        results[LOSSES] = loss_per_epoch
        results[ENTITY_TO_EMBEDDING]: Mapping[str, np.ndarray] = entity_to_embedding
        results[RELATION_TO_EMBEDDING]: Mapping[str, np.ndarray] = relation_to_embedding
        results[EVAL_SUMMARY] = eval_summary
        results[ENTITY_TO_ID] = self.entity_to_id
        results[RELATION_TO_ID] = self.rel_to_id
        results[FINAL_CONFIGURATION] = params
        return results

    def _get_train_and_test_triples(self):
        train_pos = _load_data(self.config[TRAINING_SET_PATH])

        if TEST_SET_PATH in self.config:
            test_pos = _load_data(self.config[TEST_SET_PATH])
        else:
            train_pos, test_pos = train_test_split(
                train_pos,
                test_size=self.config.get(TEST_SET_RATIO, 0.1),
                random_state=self.seed,
            )

        return self._handle_train_and_test(train_pos, test_pos)

    def _handle_train_and_test(self, train_pos, test_pos):
        all_triples = np.concatenate([train_pos, test_pos], axis=0)
        self.entity_to_id, self.rel_to_id = create_mappings(triples=all_triples)
        mapped_pos_train_triples, _, _ = create_mapped_triples(
            triples=train_pos,
            entity_to_id=self.entity_to_id,
            rel_to_id=self.rel_to_id,
        )

        mapped_pos_test_triples, _, _ = create_mapped_triples(
            triples=test_pos,
            entity_to_id=self.entity_to_id,
            rel_to_id=self.rel_to_id,
        )

        return mapped_pos_train_triples, mapped_pos_test_triples

    def _get_train_triples(self):
        train_pos = _load_data(self.config[TRAINING_SET_PATH])

        self.entity_to_id, self.rel_to_id = create_mappings(triples=train_pos)

        mapped_pos_train_triples, _, _ = create_mapped_triples(
            triples=train_pos,
            entity_to_id=self.entity_to_id,
            rel_to_id=self.rel_to_id,
        )

        return mapped_pos_train_triples


def _load_data(path: str) -> np.ndarray:
    if path.startswith('ndex:'):
        return _load_ndex(path[len('ndex:'):])

    if path.endswith('.tsv'):
        return np.loadtxt(
            fname=path,
            dtype=str,
            comments='@Comment@ Subject Predicate Object',
            delimiter='\t',
        )

    if path.endswith('.nt'):
        import rdflib
        g = rdflib.Graph()
        g.parse(path, format='nt')
        return np.array(
            [
                [str(s), str(p), str(o)]
                for s, p, o in g
            ],
            dtype=np.str,
        )

    raise ValueError('''The argument to _load_data must be one of the following:
    
    - A string path to a .tsv file containing 3 columns corresponding to subject, predicate, and object
    - A string path to a .nt RDF file serialized in N-Triples format 
    - A string NDEx network UUID prefixed by "ndex:" like in ndex:f93f402c-86d4-11e7-a10d-0ac135e8bacf
    ''')


def _load_ndex(network_uuid: str) -> np.ndarray:
    """Load a network from NDEx.

    Example network UUID: f93f402c-86d4-11e7-a10d-0ac135e8bacf
    """
    import ndex2
    ndex_client = ndex2.Ndex2()

    log.info(f'downloading {network_uuid} from ndex')
    res = ndex_client.get_network_as_cx_stream(network_uuid)
    res_json = res.json()

    triples = []

    for entry in res_json:
        for aspect, data in entry.items():
            if aspect == 'edges':
                for edge in data:
                    triples.append([
                        str(edge['s']),
                        edge.get('i', 'interacts'),
                        str(edge['t']),
                    ])

    return np.array(triples)




import torch
import numpy as np

if __name__ == '__main__':
    training_triples = torch.tensor(([[1, 1, 1], [2, 2, 2], [3, 3, 3]]))
    test_triples = torch.tensor(([[3, 3, 3], [5, 5, 5],[1,1,1]]))

    t = torch.eq(test_triples,training_triples)

    print(t)




# -*- coding: utf-8 -*-

"""Script for initializing the knowledge graph embedding models."""

from typing import Dict

import torch.optim as optim
from torch.nn import Module

from pykeen.constants import (
    ADAGRAD_OPTIMIZER_NAME, ADAM_OPTIMIZER_NAME, CONV_E_NAME, DISTMULT_NAME, ERMLP_NAME, KG_EMBEDDING_MODEL_NAME,
    LEARNING_RATE, OPTMIZER_NAME, RESCAL_NAME, SE_NAME, SGD_OPTIMIZER_NAME, TRANS_D_NAME, TRANS_E_NAME, TRANS_H_NAME,
    TRANS_R_NAME, UM_NAME,
)
from pykeen.kg_embeddings_model import (
    ConvE, DistMult, ERMLP, RESCAL, StructuredEmbedding, TransD, TransE, TransH, TransR, UnstructuredModel,
)


def get_kg_embedding_model(config: Dict) -> Module:
    """Get an instance of a knowledge graph embedding model with the given configuration."""
    model_name = config[KG_EMBEDDING_MODEL_NAME]

    if model_name == TRANS_E_NAME:
        return TransE(config=config)

    if model_name == TRANS_H_NAME:
        return TransH(config=config)

    if model_name == TRANS_D_NAME:
        return TransD(config=config)

    if model_name == TRANS_R_NAME:
        return TransR(config=config)

    if model_name == SE_NAME:
        return StructuredEmbedding(config=config)

    if model_name == UM_NAME:
        return UnstructuredModel(config=config)

    if model_name == DISTMULT_NAME:
        return DistMult(config=config)

    if model_name == ERMLP_NAME:
        return ERMLP(config=config)

    if model_name == RESCAL_NAME:
        return RESCAL(config=config)

    if model_name == CONV_E_NAME:
        return ConvE(config=config)

    raise ValueError(f'Invalid KGE model name: {model_name}')


def get_optimizer(config, kg_embedding_model):
    """

    :param config:
    :return:
    """
    optimizer_name = config[OPTMIZER_NAME]

    parameters = filter(lambda p: p.requires_grad, kg_embedding_model.parameters())

    if optimizer_name == SGD_OPTIMIZER_NAME:
        return optim.SGD(parameters, lr=config[LEARNING_RATE])

    if optimizer_name == ADAGRAD_OPTIMIZER_NAME:
        optim.Adagrad(parameters, lr=config[LEARNING_RATE])

    if optimizer_name == ADAM_OPTIMIZER_NAME:
        optim.Adam(parameters, lr=config[LEARNING_RATE])




from itertools import product
import numpy as np
import torch

a = [1,2,3]
c = list(product(a, a))

# print(np.array(c))

d = torch.tensor([100,20,50])

_, e = torch.sort(d)
e = e.cpu().numpy()

print(e)


# -*- coding: utf-8 -*-
import logging
import os
from urllib.request import urlretrieve

from corpus_reader.AbstractReader import AbstractReader
from keen.utilities import KG_EMBEDDINGS_PIPELINE_DIR, WROC, WROC_URL

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class WROCReader(AbstractReader):

    def transorm_corpus_to_id_format(self, temp_corpus, corpus_path):
        """

        :param temp_corpus:
        :param corpus_path:
        :return:
        """

        with open(temp_corpus, 'r', encoding='utf-8') as f1, open(corpus_path, 'w', encoding='utf-8') as f2:
            lines = f1.readlines()
            f2.write('@Comment@ Subject Predicate Object\n')
            for line in lines:
                parts = line.strip().split(' ')
                parts = [p for p in parts if p != '' and p != '.']
                assert len(parts) == 3
                f2.write('\t'.join(parts) + '\n')

        os.remove(temp_corpus)

    def retreive_knowledge_graph(self, enforce_download=False):
        # Implementation based on https://github.com/bio2bel/hmdb/blob/master/src/bio2bel_hmdb/parser.py
        """

        :param enforce_download:
        :return:
        """
        data_dir = os.path.join(KG_EMBEDDINGS_PIPELINE_DIR, WROC)
        os.makedirs(data_dir, exist_ok=True)

        corpus_path = os.path.join(data_dir, "walking_rdf_and_owl.csv")

        if os.path.exists(corpus_path) and not enforce_download:
            log.info('Used existing corpus at %s', corpus_path)
        else:
            log.info('Download corpus')
            path, _ = urlretrieve(WROC_URL)

            self.transorm_corpus_to_id_format(temp_corpus=path, corpus_path=corpus_path)

        return corpus_path









# -*- coding: utf-8 -*-

"""Tests for `utils`."""

import unittest

from pykeen.constants import VERSION


class SimpleTest(unittest.TestCase):
    """Simple sanity tests."""

    def test_version(self):
        """Check the version string can be loaded."""
        self.assertIsInstance(VERSION, str)



# -*- coding: utf-8 -*-
import logging
import os
from urllib.request import urlretrieve
import zipfile


from keen.utilities import KG_EMBEDDINGS_PIPELINE_DIR, WN_18, WN18_URL

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)
from corpus_reader.AbstractReader import AbstractReader


class WN18Reader(AbstractReader):

    def transorm_corpus_to_id_format(self, temp_corpus, corpus_path):
        """

        :param temp_corpus:
        :param corpus_path:
        :return:
        """

        data_dir = os.path.join(KG_EMBEDDINGS_PIPELINE_DIR, WN_18)
        os.makedirs(data_dir, exist_ok=True)
        zip_ref = zipfile.ZipFile(temp_corpus, 'r')
        zip_ref.extractall(data_dir)
        zip_ref.close()

        train_temp = os.path.join(data_dir,'WN18','train.txt')
        test_temp = os.path.join(data_dir,'test.txt')
        valid_temp = os.path.join(data_dir, 'valid.txt')

        train = os.path.join(data_dir,'wn_18_train.tsv')

        with open(train_temp, 'r', encoding='utf-8') as f1, open(train, 'w', encoding='utf-8') as f2:
            lines = f1.readlines()
            f2.write('@Comment@ Subject Predicate Object\n')
            for line in lines:
                parts = line.strip().split('\t')
                subject = parts[0]
                object = parts[1]
                predicate = parts[2]
                assert len(parts) == 3
                parts = [subject,predicate,object]
                f2.write('\t'.join(parts) + '\n')

        os.remove(temp_corpus)

    def retreive_knowledge_graph(self, enforce_download=False):
        """

        :return:
        """
        data_dir = os.path.join(KG_EMBEDDINGS_PIPELINE_DIR, WN_18)
        os.makedirs(data_dir, exist_ok=True)

        corpus_path = os.path.join(data_dir, "wn_18_train.tsv")

        if os.path.exists(corpus_path) and not enforce_download:
            log.info('Used existing corpus at %s', corpus_path)
        else:
            log.info('Download corpus')
            path, _ = urlretrieve(WN18_URL)
            self.transorm_corpus_to_id_format(temp_corpus=path, corpus_path=corpus_path)

        return corpus_path


# -*- coding: utf-8 -*-

"""Implementations of various knowledge graph embedding models."""

from pykeen.kg_embeddings_model.conv_e import ConvE  # noqa: F401
from pykeen.kg_embeddings_model.distmult import DistMult  # noqa: F401
from pykeen.kg_embeddings_model.ermlp import ERMLP  # noqa: F401
from pykeen.kg_embeddings_model.rescal import RESCAL  # noqa: F401
from pykeen.kg_embeddings_model.structured_embedding import StructuredEmbedding  # noqa: F401
from pykeen.kg_embeddings_model.trans_d import TransD  # noqa: F401
from pykeen.kg_embeddings_model.trans_e import TransE  # noqa: F401
from pykeen.kg_embeddings_model.trans_h import TransH  # noqa: F401
from pykeen.kg_embeddings_model.trans_r import TransR  # noqa: F401
from pykeen.kg_embeddings_model.unstructured_model import UnstructuredModel  # noqa: F401



# -*- coding: utf-8 -*-

"""Implementation of TransH."""

import numpy as np
import torch
import torch.autograd
import torch.nn as nn

from pykeen.constants import *

__all__ = ['TransH']


class TransH(nn.Module):

    def __init__(self, config):
        super(TransH, self).__init__()
        self.model_name = TRANS_H_NAME
        self.device = torch.device(
            'cuda:0' if torch.cuda.is_available() and config[PREFERRED_DEVICE] == GPU else CPU)
        self.use_cuda = True if torch.cuda.is_available() and config[PREFERRED_DEVICE] == GPU else False
        self.num_entities = config[NUM_ENTITIES]
        self.num_relations = config[NUM_RELATIONS]
        self.embedding_dim = config[EMBEDDING_DIM]
        self.config = config
        margin_loss = config[MARGIN_LOSS]

        # A simple lookup table that stores embeddings of a fixed dictionary and size
        self.entity_embeddings = nn.Embedding(self.num_entities, self.embedding_dim)
        self.relation_embeddings = nn.Embedding(self.num_relations, self.embedding_dim)
        self.normal_vector_embeddings = nn.Embedding(self.num_relations, self.embedding_dim)
        self.margin_loss = margin_loss
        self.weightning_soft_constraint = config[WEIGHT_SOFT_CONSTRAINT_TRANS_H]
        self.criterion = nn.MarginRankingLoss(margin=self.margin_loss, size_average=False)
        self.epsilon = 0.05
        self.scoring_fct_norm = config[SCORING_FUNCTION_NORM]

    def _initialize(self):
        # TODO: Add initialization
        pass

    def project_to_hyperplane(self, entity_embs, normal_vec_embs):
        """

        :param entity_embs: Embeddings of entities with shape batch_size x 1 x embedding_dimension
        :param normal_vec_embs: Normal vectors with shape batch_size x 1 x embedding_dimension
        :return: Projected entities of shape batch_size x embedding_dim
        """

        scaling_factors = torch.sum(normal_vec_embs * entity_embs, dim=-1).unsqueeze(1)
        heads_projected_on_normal_vecs = scaling_factors * normal_vec_embs
        projections = (entity_embs - heads_projected_on_normal_vecs).view(-1, self.embedding_dim)

        return projections

    def _compute_scores(self, h_embs, r_embs, t_embs):
        """

        :param h_embs: Head embeddings of shape batch_size x embedding_dim
        :param r_embs: Relation emebddings of shape of batch_size x embedding_dim
        :param t_embs: Tail embeddings of shape batch_size x embedding_dim
        :return: Scores of shape batch_size
        """

        # Add the vector element wise
        sum_res = h_embs + r_embs - t_embs
        scores = torch.norm(sum_res, dim=1, p=self.scoring_fct_norm).view(size=(-1,))

        return scores

    def compute_soft_constraint_loss(self, batch_entities, batch_relations):
        """
        Compute the soft constraints.
        :return:
        """

        entity_embs = self.entity_embeddings(batch_entities)
        normal_vector_embeddings = self.normal_vector_embeddings(batch_relations)
        relation_embeddings = self.relation_embeddings(batch_relations)

        norm_of_entities = torch.norm(entity_embs, p=2, dim=1)
        square_norms_entities = torch.mul(norm_of_entities, norm_of_entities)
        entity_constraint = square_norms_entities - self.num_entities * 1.
        entity_constraint = torch.abs(entity_constraint)
        entity_constraint = torch.sum(entity_constraint)

        orthogonalty_constraint_numerator = torch.mul(normal_vector_embeddings, relation_embeddings)
        orthogonalty_constraint_numerator = torch.sum(orthogonalty_constraint_numerator, dim=1)
        orthogonalty_constraint_numerator = torch.mul(orthogonalty_constraint_numerator,
                                                      orthogonalty_constraint_numerator)

        orthogonalty_constraint_denominator = torch.norm(relation_embeddings, p=2, dim=1)
        orthogonalty_constraint_denominator = torch.mul(orthogonalty_constraint_denominator,
                                                        orthogonalty_constraint_denominator)

        orthogonalty_constraint = (orthogonalty_constraint_numerator / orthogonalty_constraint_denominator) - \
                                  (self.num_relations * self.epsilon)
        orthogonalty_constraint = torch.abs(orthogonalty_constraint)
        orthogonalty_constraint = torch.sum(orthogonalty_constraint)

        soft_constraints_loss = self.weightning_soft_constraint * (entity_constraint + orthogonalty_constraint)

        return soft_constraints_loss

    def compute_loss(self, pos_scores, neg_scores, batch_entities, batch_relations):
        """

        :param pos_scores:
        :param neg_scores:
        :return:
        """

        pos_scores = torch.tensor(pos_scores, dtype=torch.float, device=self.device)
        neg_scores = torch.tensor(neg_scores, dtype=torch.float, device=self.device)

        # y == -1 indicates that second input to criterion should get a larger loss
        y = np.repeat([-1], repeats=pos_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)
        margin_ranking_loss = self.criterion(pos_scores, neg_scores, y)
        soft_constraint_loss = self.compute_soft_constraint_loss(batch_entities, batch_relations)

        loss = margin_ranking_loss + soft_constraint_loss

        return loss

    def predict(self, triples):
        """

        :param head:
        :param relation:
        :param tail:
        :return:
        """

        heads = triples[:, 0:1]
        relations = triples[:, 1:2]
        tails = triples[:, 2:3]

        head_embs = self.entity_embeddings(heads)
        tail_embs = self.entity_embeddings(tails)
        relation_embs = self.relation_embeddings(relations).view(-1, self.embedding_dim)
        normal_vec_embs = self.normal_vector_embeddings(relations)

        head_embs_projected = self.project_to_hyperplane(entity_embs=head_embs, normal_vec_embs=normal_vec_embs)
        tail_embs_projected = self.project_to_hyperplane(entity_embs=tail_embs, normal_vec_embs=normal_vec_embs)

        scores = self._compute_scores(h_embs=head_embs_projected, r_embs=relation_embs, t_embs=tail_embs_projected)

        return scores.detach().cpu().numpy()

    def forward(self, batch_positives, batch_negatives):
        """

        :param batch_positives:
        :param batch_negatives:
        :return:
        """

        # Normalise the normal vectors by their l2 norms
        norms = torch.norm(self.normal_vector_embeddings.weight, p=2, dim=1).data
        self.normal_vector_embeddings.weight.data = self.normal_vector_embeddings.weight.data.div(
            norms.view(self.num_relations, 1).expand_as(self.normal_vector_embeddings.weight))

        pos_heads = batch_positives[:, 0:1]
        pos_rels = batch_positives[:, 1:2]
        pos_tails = batch_positives[:, 2:3]

        neg_heads = batch_negatives[:, 0:1]
        neg_rels = batch_negatives[:, 1:2]
        neg_tails = batch_negatives[:, 2:3]

        # Shape: (batch_size, 1, embedding_dimension)
        pos_head_embs = self.entity_embeddings(pos_heads)
        # Reshape relation embeddings to the same shape of the projected entities
        pos_rel_embs = self.relation_embeddings(pos_rels).view(-1, self.embedding_dim)
        pos_tail_embs = self.entity_embeddings(pos_tails)
        pos_normal_embs = self.normal_vector_embeddings(pos_rels)

        neg_head_embs = self.entity_embeddings(neg_heads)
        # Reshape relation embeddings to the same shape of the projected entities
        neg_rel_embs = self.relation_embeddings(neg_rels).view(-1, self.embedding_dim)
        neg_tail_embs = self.entity_embeddings(neg_tails)
        neg_normal_embs = self.normal_vector_embeddings(neg_rels)

        projected_heads_pos = self.project_to_hyperplane(entity_embs=pos_head_embs, normal_vec_embs=pos_normal_embs)
        projected_tails_pos = self.project_to_hyperplane(entity_embs=pos_tail_embs, normal_vec_embs=pos_normal_embs)

        projected_heads_neg = self.project_to_hyperplane(entity_embs=neg_head_embs, normal_vec_embs=neg_normal_embs)
        projected_tails_neg = self.project_to_hyperplane(entity_embs=neg_tail_embs, normal_vec_embs=neg_normal_embs)

        pos_scores = self._compute_scores(h_embs=projected_heads_pos, r_embs=pos_rel_embs, t_embs=projected_tails_pos)
        neg_scores = self._compute_scores(h_embs=projected_heads_neg, r_embs=neg_rel_embs, t_embs=projected_tails_neg)

        batch_entities = torch.cat([pos_heads.view(-1), pos_tails.view(-1), neg_heads.view(-1), neg_tails.view(-1)])
        batch_entities = torch.unique(batch_entities.cpu())
        batch_relations = torch.unique(torch.cat([pos_rels.view(-1), neg_rels.view(-1)]).cpu())

        if self.use_cuda:
            batch_entities = batch_entities.cuda()
            batch_relations = batch_relations.cuda()

        loss = self.compute_loss(pos_scores=pos_scores, neg_scores=neg_scores, batch_entities=batch_entities,
                                 batch_relations=batch_relations)
        return loss



import numpy as np
if __name__ == '__main__':
    path = '/Users/mehdi/PycharmProjects/kg_embeddings_pipeline/corpora/fb15k/fb_15k_test.tsv'

    with open(path) as f:
        lines = f.readlines()

        for l in lines:
            p = l.split('\t')
            if len(p) >3:
                print(p)

        pos_triples = np.loadtxt(fname=path, dtype=str, comments='@Comment@ Subject Predicate Object')#, delimiter='\t')

        print(pos_triples)


import json
import numpy as np

if __name__ == '__main__':
    input_path = '/Users/mehdi/PycharmProjects/BioKEEN/data/out/2018-11-20_12:52:31/entity_to_id.json'
    out = '/Users/mehdi/PycharmProjects/BioKEEN/data/out/2018-11-20_12:52:31/entities.tsv'
    with open(input_path) as json_data:
        entity_to_id = json.load(json_data)

    entities = np.array(list(entity_to_id.keys()),dtype=str)
    entities = np.reshape(entities,newshape=(-1,1))

    np.savetxt(fname=out,X=entities,fmt='%s')

    print(entities)


# -*- coding: utf-8 -*-
import os
import pickle
import sys
import time

from keen.utilities import KG_EMBEDDINGS_PIPELINE_DIR, ENTITY_TO_EMBEDDINGS, EVAL_RESULTS, CSQA_WIKIDATA

w_dir = os.path.dirname(os.getcwd())
sys.path.append(w_dir)

import click
import yaml
from keen.utilities.pipeline import Pipeline


@click.command()
@click.option('-cfg_path', help='path to config file', required=True)
def main(cfg_path):
    with open(cfg_path, 'r') as ymlfile:
        cfg = yaml.load(ymlfile)

    pipeline = Pipeline(config=cfg)

    trained_kg_model, eval_summary, entity_to_embedding, relation_to_embedding = pipeline._start_pipeline(
        learning_rate=0.01, num_epochs=1,
        ratio_of_neg_triples=0.5,
        batch_size=None,
        ratio_test_data=1/5, seed=2)

    current_time = time.strftime("%H:%M:%S")
    current_date = time.strftime("%d/%m/%Y").replace('/', '-')
    entity_to_embedding_out = os.path.join(KG_EMBEDDINGS_PIPELINE_DIR, CSQA_WIKIDATA, 'output', current_date,
                                           current_time)
    os.makedirs(entity_to_embedding_out, exist_ok=True)

    eval_results_out = os.path.join(KG_EMBEDDINGS_PIPELINE_DIR, CSQA_WIKIDATA, 'output', current_date, current_time)

    os.makedirs(eval_results_out, exist_ok=True)

    entity_to_embedding_out = os.path.join(entity_to_embedding_out, ENTITY_TO_EMBEDDINGS + '.pkl')
    eval_results_out = os.path.join(eval_results_out, EVAL_RESULTS + '.pkl')

    with open(entity_to_embedding_out, 'wb') as handle:
        pickle.dump(entity_to_embedding, handle, protocol=pickle.HIGHEST_PROTOCOL)

    with open(eval_results_out, 'wb') as handle:
        pickle.dump(eval_summary, handle, protocol=pickle.HIGHEST_PROTOCOL)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""PyKEEN's command line interface."""

import json
import os
from collections import OrderedDict
from typing import Dict

import click

from pykeen.constants import (
    EXECUTION_MODE, FILTER_NEG_TRIPLES, HPO_ITERS_ERROR_MSG, HPO_ITERS_PRINT_MSG, HPO_ITERS_PROMPT_MSG, HPO_MODE,
    NUM_OF_HPO_ITERS, OUTPUT_DIREC, PREFERRED_DEVICE, PYKEEN, SEED, SEED_ERROR_MSG, SEED_PRINT_MSG, SEED_PROMPT_MSG,
    TEST_FILE_ERROR_MSG, TEST_FILE_PROMPT_MSG, TEST_SET_PATH, TEST_SET_RATIO, TRAINING_FILE_ERROR_MSG,
    TRAINING_FILE_PROMPT_MSG, TRAINING_MODE, TRAINING_SET_PATH,
)
from pykeen.predict import start_predictions_pipeline
from pykeen.run import run
from pykeen.utilities.cli_utils.cli_print_msg_helper import (
    print_ask_for_evlauation_message, print_execution_mode_message, print_filter_negative_triples_message, print_intro,
    print_model_selection_message, print_random_seed_message, print_section_divider, print_test_ratio_message,
    print_test_set_message, print_training_set_message, print_welcome_message,
)
from pykeen.utilities.cli_utils.cli_query_helper import (
    ask_for_evaluation, ask_for_filtering_of_negatives, ask_for_test_set, get_input_path, query_output_directory,
    select_embedding_model, select_integer_value, select_keen_execution_mode, select_preferred_device,
    select_ratio_for_test_set,
)
from pykeen.utilities.cli_utils.dicts import MODEL_HPO_CONFIG_FUNCS, MODEL_TRAINING_CONFIG_FUNCS
from pykeen.utilities.cli_utils.utils import summarize_results


def _configure_training_pipeline(model_name: str):
    model_config_func = MODEL_TRAINING_CONFIG_FUNCS.get(model_name)
    if model_config_func is None:
        raise KeyError(f'invalid model given: {model_name}')
    config = model_config_func(model_name)
    if config is None:
        raise NotImplementedError(f'{model_name} has not yet been implemented')
    return config


def _configure_hpo_pipeline(model_name: str):
    model_config_func = MODEL_HPO_CONFIG_FUNCS.get(model_name)
    if model_config_func is None:
        raise KeyError(f'invalid model given: {model_name}')
    config = model_config_func(model_name)
    if config is None:
        raise NotImplementedError(f'{model_name} has not yet been implemented')
    return config


def prompt_evaluation_parameters(config: Dict) -> None:
    """Prompt the user for evaluation parameters absed on the execution mode."""
    if config[EXECUTION_MODE] == TRAINING_MODE:
        # Step 1: Ask whether to evaluate the model
        print_ask_for_evlauation_message()
        is_evaluation_mode = ask_for_evaluation()
        print_section_divider()
    else:
        is_evaluation_mode = True

    # Step 2: Specify test set, if is_evaluation_mode==True
    if is_evaluation_mode:
        print_test_set_message()
        provide_test_set = ask_for_test_set()
        print_section_divider()

        if provide_test_set:
            test_set_path = get_input_path(
                prompt_msg=TEST_FILE_PROMPT_MSG,
                error_msg=TEST_FILE_ERROR_MSG,
            )
            config[TEST_SET_PATH] = test_set_path
        else:
            print_test_ratio_message()
            test_set_ratio = select_ratio_for_test_set()
            config[TEST_SET_RATIO] = test_set_ratio

        print_section_divider()

        # Ask whether to use filtered negative triples
        print_filter_negative_triples_message()
        filter_negative_triples = ask_for_filtering_of_negatives()
        config[FILTER_NEG_TRIPLES] = filter_negative_triples
        print_section_divider()


def prompt_training_file(config: Dict) -> None:
    """Prompt the user for a training file."""
    print_training_set_message()
    config[TRAINING_SET_PATH] = get_input_path(
        prompt_msg=TRAINING_FILE_PROMPT_MSG,
        error_msg=TRAINING_FILE_ERROR_MSG,
        is_dataset=True,
    )


def prompt_execution_mode(config: Dict, lib_name: str = PYKEEN) -> None:
    """Prompt the user for the execution mode."""
    print_execution_mode_message()
    config[EXECUTION_MODE] = select_keen_execution_mode(lib_name=lib_name)


def prompt_embedding_model() -> str:
    """Prompt the user to select an embedding model."""
    print_model_selection_message()
    model_name = select_embedding_model()
    print_section_divider()
    return model_name


def prompt_execution_parameters(config: Dict, model_name: str) -> None:
    """Prompt the user for execution mode parameters."""
    pykeen_exec_mode = config[EXECUTION_MODE]

    if pykeen_exec_mode == TRAINING_MODE:
        config.update(_configure_training_pipeline(model_name))

    elif pykeen_exec_mode == HPO_MODE:
        config.update(_configure_hpo_pipeline(model_name))

        # Query number of HPO iterations
        hpo_iter = select_integer_value(
            print_msg=HPO_ITERS_PRINT_MSG,
            prompt_msg=HPO_ITERS_PROMPT_MSG,
            error_msg=HPO_ITERS_ERROR_MSG,
        )
        config[NUM_OF_HPO_ITERS] = hpo_iter
        print_section_divider()


def prompt_random_seed(config):
    """Query random seed."""
    print_random_seed_message()
    config[SEED] = select_integer_value(
        print_msg=SEED_PRINT_MSG,
        prompt_msg=SEED_PROMPT_MSG,
        error_msg=SEED_ERROR_MSG,
    )


def prompt_device(config: Dict) -> None:
    """Prompt the user for their preferred evaluation device."""
    config[PREFERRED_DEVICE] = select_preferred_device()


def prompt_output_directory(config: Dict) -> None:
    """Prompt the user for the output directory."""
    config[OUTPUT_DIREC] = query_output_directory()


def prompt_config() -> Dict:
    """Prompt the user for the run configuration."""
    cfg = OrderedDict()

    # Step 1: Welcome + Intro
    print_welcome_message()
    print_section_divider()

    print_intro()
    print_section_divider()

    # Step 2: Ask for training file
    prompt_training_file(cfg)
    print_section_divider()

    # Step 3: Ask for execution mode
    prompt_execution_mode(cfg)
    print_section_divider()

    # Step 4: Ask for model
    model_name = prompt_embedding_model()
    print_section_divider()

    # Step 5: Query parameters depending on the selected execution mode
    prompt_execution_parameters(cfg, model_name=model_name)
    print_section_divider()

    # Step 5.5: Prompt for evaluation parameters depending on the selected execution mode
    prompt_evaluation_parameters(cfg)

    # Step 6: Please select a random seed
    prompt_random_seed(cfg)
    print_section_divider()

    # Step 7: Query device to train on
    prompt_device(cfg)
    print_section_divider()

    # Step 8: Define output directory
    prompt_output_directory(cfg)
    print_section_divider()

    return cfg


@click.command()
@click.option('-c', '--config', type=click.File(), help='A PyKEEN JSON configuration file')
def main(config):
    """PyKEEN: A software for training and evaluating knowledge graph embeddings."""
    if config is not None:
        config = json.load(config)
    else:
        config = prompt_config()

    run(config)


@click.command()
@click.option('-m', '--model-directory', type=click.Path(file_okay=False, dir_okay=True))
@click.option('-d', '--data-directory', type=click.Path(file_okay=False, dir_okay=True))
def predict(model_directory: str, data_directory: str):
    """Predict new links based on trained model."""
    start_predictions_pipeline(model_directory, data_directory)


@click.command()
@click.option('-d', '--directory', type=click.Path(file_okay=False, dir_okay=True), default=os.getcwd())
@click.option('-o', '--output', type=click.File('w'))
def summarize(directory: str, output):
    """Summarize contents of training and evaluation."""
    summarize_results(directory, output)


if __name__ == '__main__':
    main()




import matplotlib
import matplotlib.pyplot as plt
matplotlib.matplotlib_fname()
import numpy as np
import os
import json

def plot_losses(losses, output_direc):
    out_path = os.path.join(output_direc, 'losses.png')
    epochs = np.arange(len(losses))
    plt.title(r'Loss Per Epoch')
    plt.xlabel('epoch')
    plt.ylabel('loss')

    plt.plot(epochs, losses)
    plt.savefig(out_path)

if __name__ == '__main__':
    input_path = '/Users/mehdi/PycharmProjects/BioKEEN/data/out/2018-11-13_17:16:42/losses.json'
    output_direc = '/Users/mehdi/PycharmProjects/BioKEEN/data/out/2018-11-13_17:16:42'

    with open(input_path) as json_data:
        losses = json.load(json_data)

    plot_losses(losses=losses, output_direc=output_direc)




import numpy as np

A = np.array([[1,4],[2,5],[3,6]])
B = np.array([[1,4],[3,6],[7,8]])


nrows, ncols = A.shape
dtype={'names':['f{}'.format(i) for i in range(ncols)],
       'formats':ncols * [A.dtype]}

C = np.intersect1d(A.view(dtype), B.view(dtype))

# This last bit is optional if you're okay with "C" being a structured array...
C = C.view(A.dtype).reshape(-1, ncols)

print(C)

d = np.setdiff1d(B, A)

print(d)


# -*- coding: utf-8 -*-

"""Implementation of DistMult."""

import numpy as np
import torch
import torch.autograd
import torch.nn as nn

from pykeen.constants import *

__all__ = ['DistMult']


class DistMult(nn.Module):
    """An implementation of DistMult."""

    def __init__(self, config):
        super(DistMult, self).__init__()
        self.model_name = DISTMULT_NAME
        # A simple lookup table that stores embeddings of a fixed dictionary and size
        self.num_entities = config[NUM_ENTITIES]
        self.num_relations = config[NUM_RELATIONS]
        self.embedding_dim = config[EMBEDDING_DIM]

        self.device = torch.device(
            'cuda:0' if torch.cuda.is_available() and config[PREFERRED_DEVICE] == GPU else CPU)

        # self.l_p_norm_entities = config[NORM_FOR_NORMALIZATION_OF_ENTITIES]
        self.entity_embeddings = nn.Embedding(self.num_entities, self.embedding_dim)
        self.relation_embeddings = nn.Embedding(self.num_relations, self.embedding_dim)

        self.margin_loss = config[MARGIN_LOSS]
        self.criterion = nn.MarginRankingLoss(margin=self.margin_loss, size_average=True)

        self._initialize()

    def _initialize(self):
        """

        :return:
        """
        lower_bound = -6 / np.sqrt(self.embedding_dim)
        upper_bound = 6 / np.sqrt(self.embedding_dim)
        nn.init.uniform_(self.entity_embeddings.weight.data, a=lower_bound, b=upper_bound)
        nn.init.uniform_(self.relation_embeddings.weight.data, a=lower_bound, b=upper_bound)

        norms = torch.norm(self.relation_embeddings.weight, p=2, dim=1).data
        self.relation_embeddings.weight.data = self.relation_embeddings.weight.data.div(
            norms.view(self.num_relations, 1).expand_as(self.relation_embeddings.weight))

    def _compute_loss(self, pos_scores, neg_scores):
        """

        :param pos_scores:
        :param neg_scores:
        :return:
        """

        # Choose y = -1 since a smaller score is better.
        # In TransE for exampel the scores represent distances
        y = np.repeat([-1], repeats=pos_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the psotive and negative triples
        pos_scores = torch.tensor(pos_scores, dtype=torch.float, device=self.device)
        neg_scores = torch.tensor(neg_scores, dtype=torch.float, device=self.device)
        # neg_scores_temp = 1 * torch.tensor(neg_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(pos_scores, neg_scores, y)

        return loss

    def _compute_scores(self, h_embs, r_embs, t_embs):
        """

        :param h_embs:
        :param r_embs:
        :param t_embs:
        :return:
        """
        scores = - torch.sum(h_embs * r_embs * t_embs, dim=1)

        return scores

    def predict(self, triples):
        """

        :param head:
        :param relation:
        :param tail:
        :return:
        """
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)

        heads = triples[:, 0:1]
        relations = triples[:, 1:2]
        tails = triples[:, 2:3]

        head_embs = self.entity_embeddings(heads).view(-1, self.embedding_dim)
        relation_embs = self.relation_embeddings(relations).view(-1, self.embedding_dim)
        tail_embs = self.entity_embeddings(tails).view(-1, self.embedding_dim)

        scores = self._compute_scores(h_embs=head_embs, r_embs=relation_embs, t_embs=tail_embs)

        return scores.detach().cpu().numpy()

    def forward(self, batch_positives, batch_negatives):
        """

        :param batch_positives:
        :param batch_negatives:
        :return:
        """

        pos_heads = batch_positives[:, 0:1]
        pos_relations = batch_positives[:, 1:2]
        pos_tails = batch_positives[:, 2:3]

        neg_heads = batch_negatives[:, 0:1]
        neg_relations = batch_negatives[:, 1:2]
        neg_tails = batch_negatives[:, 2:3]

        pos_h_embs = self.entity_embeddings(pos_heads).view(-1, self.embedding_dim)
        pos_r_embs = self.relation_embeddings(pos_relations).view(-1, self.embedding_dim)
        pos_t_embs = self.entity_embeddings(pos_tails).view(-1, self.embedding_dim)

        neg_h_embs = self.entity_embeddings(neg_heads).view(-1, self.embedding_dim)
        neg_r_embs = self.relation_embeddings(neg_relations).view(-1, self.embedding_dim)
        neg_t_embs = self.entity_embeddings(neg_tails).view(-1, self.embedding_dim)

        pos_scores = self._compute_scores(h_embs=pos_h_embs, r_embs=pos_r_embs, t_embs=pos_t_embs)
        neg_scores = self._compute_scores(h_embs=neg_h_embs, r_embs=neg_r_embs, t_embs=neg_t_embs)

        loss = self._compute_loss(pos_scores=pos_scores, neg_scores=neg_scores)

        return loss




import os
import getpass


if __name__ == '__main__':
    base_dir = os.environ.get('KG',os.path.join('/data', getpass.getuser(), 'kg_embeddings_pipeline'))
    os.makedirs(base_dir, exist_ok=True)

    print(base_dir)
    print(getpass.getuser())




import torch

if __name__ == '__main__':
        h_embs = torch.tensor([[1,2],[3,4]])
        r_embs = torch.tensor([[2,1],[0,1]])
        t_embs = torch.tensor([[1, 2], [3, 4]])

        intermediates = torch.mul(h_embs, r_embs)
        scores = torch.einsum('nd,nd->n', [intermediates, t_embs])

        print(scores)



# -*- coding: utf-8 -*-

"""Utilities for hyper-parameter optimization."""

import random
from abc import ABC, abstractmethod
from typing import Any, Iterable, Mapping

__all__ = ['AbstractHPOptimizer']


class AbstractHPOptimizer(ABC):
    """An abstract class from which all hyper-parameter optimizers should inherit."""

    @abstractmethod
    def optimize_hyperparams(self, config, path_to_kg, device, seed):
        """Run the optimizer."""

    @staticmethod
    def _sample_parameter_value(parameter_values: Mapping[int, Iterable[Any]]) -> Mapping[int, Any]:
        """Randomly subsample a dictionary whose values are iterable."""
        return {
            parameter: (
                random.choice(values)
                if isinstance(values, list) else
                values
            )
            for parameter, values in parameter_values.items()
        }



# -*- coding: utf-8 -*-

"""Implementation of UM."""

import logging

import numpy as np
import torch
import torch.autograd
import torch.nn as nn

from pykeen.constants import *

__all__ = ['UnstructuredModel']

log = logging.getLogger(__name__)


class UnstructuredModel(nn.Module):

    def __init__(self, config):
        super(UnstructuredModel, self).__init__()
        self.model_name = UM_NAME
        # A simple lookup table that stores embeddings of a fixed dictionary and size
        self.num_entities = config[NUM_ENTITIES]
        self.num_relations = config[NUM_RELATIONS]
        self.embedding_dim = config[EMBEDDING_DIM]

        self.device = torch.device(
            'cuda:0' if torch.cuda.is_available() and config[PREFERRED_DEVICE] == GPU else CPU)

        self.l_p_norm_entities = config[NORM_FOR_NORMALIZATION_OF_ENTITIES]
        self.scoring_fct_norm = config[SCORING_FUNCTION_NORM]
        self.entity_embeddings = nn.Embedding(self.num_entities, self.embedding_dim)

        self.margin_loss = config[MARGIN_LOSS]
        self.criterion = nn.MarginRankingLoss(margin=self.margin_loss, size_average=True)

        self._initialize()

    def _initialize(self):
        """

        :return:
        """
        lower_bound = -6 / np.sqrt(self.embedding_dim)
        upper_bound = 6 / np.sqrt(self.embedding_dim)
        nn.init.uniform_(self.entity_embeddings.weight.data, a=lower_bound, b=upper_bound)

    def _compute_loss(self, pos_scores, neg_scores):
        """

        :param pos_scores:
        :param neg_scores:
        :return:
        """

        y = np.repeat([-1], repeats=pos_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the psotive and negative triples
        pos_scores = torch.tensor(pos_scores, dtype=torch.float, device=self.device)
        neg_scores = torch.tensor(neg_scores, dtype=torch.float, device=self.device)
        # neg_scores_temp = 1 * torch.tensor(neg_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(pos_scores, neg_scores, y)

        return loss

    def _compute_scores(self, h_embs, t_embs):
        """

        :param h_embs:
        :param r_embs:
        :param t_embs:
        :return:
        """

        # Add the vector element wise
        sum_res = h_embs - t_embs
        distances = torch.norm(sum_res, dim=1, p=self.scoring_fct_norm).view(size=(-1,))
        distances = distances ** 2
        return distances

    def predict(self, triples):
        """

        :param head:
        :param relation:
        :param tail:
        :return:
        """
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)

        heads = triples[:, 0:1]
        relations = triples[:, 1:2]
        tails = triples[:, 2:3]

        head_embs = self.entity_embeddings(heads).view(-1, self.embedding_dim)
        tail_embs = self.entity_embeddings(tails).view(-1, self.embedding_dim)

        scores = self._compute_scores(h_embs=head_embs, t_embs=tail_embs)

        return scores.detach().cpu().numpy()

    def forward(self, batch_positives, batch_negatives):
        """

        :param batch_positives:
        :param batch_negatives:
        :return:
        """

        # Normalise embeddings of entities
        pos_heads = batch_positives[:, 0:1]
        pos_tails = batch_positives[:, 2:3]

        neg_heads = batch_negatives[:, 0:1]
        neg_tails = batch_negatives[:, 2:3]

        pos_h_embs = self.entity_embeddings(pos_heads).view(-1, self.embedding_dim)
        pos_t_embs = self.entity_embeddings(pos_tails).view(-1, self.embedding_dim)

        neg_h_embs = self.entity_embeddings(neg_heads).view(-1, self.embedding_dim)
        neg_t_embs = self.entity_embeddings(neg_tails).view(-1, self.embedding_dim)

        pos_scores = self._compute_scores(h_embs=pos_h_embs, t_embs=pos_t_embs)
        neg_scores = self._compute_scores(h_embs=neg_h_embs, t_embs=neg_t_embs)

        loss = self._compute_loss(pos_scores=pos_scores, neg_scores=neg_scores)

        return loss




import torch


def project():
    pass

if __name__ == '__main__':
    heads = torch.tensor([[1.,1.],[2.,2.]]).unsqueeze(1)
    normal_vecs = torch.tensor([[1., 1.], [2., 2.]]).unsqueeze(1)

    scaling_factors = torch.sum(normal_vecs*heads, dim=-1).unsqueeze(1)

    print(scaling_factors.shape)
    exit(0)

    heads_projected_on_normal_vecs = scaling_factors * normal_vecs
    projections = (heads - heads_projected_on_normal_vecs).view(-1,2)

    print(projections)
    print()

    energy_vecs = heads.view(-1,2) - projections

    print(energy_vecs)
    print()

    scores = torch.norm(energy_vecs, dim=1, p=2).view(size=(-1,))

    print(scores)



# -*- coding: utf-8 -*-

"""A hyper-parameter optimizer that uses random search."""

import random
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from torch.nn import Module

from pykeen.constants import *
from pykeen.hyper_parameter_optimizer.abstract_hyper_params_optimizer import AbstractHPOptimizer
from pykeen.utilities.evaluation_utils.metrics_computations import compute_metric_results
from pykeen.utilities.initialization_utils.module_initialization_utils import get_kg_embedding_model
from pykeen.utilities.train_utils import train_model

__all__ = ['RandomSearchHPO']

OptimizeResult = Tuple[Module, List[float], Any, Any, Any, Any]


class RandomSearchHPO(AbstractHPOptimizer):
    """A hyper-parameter optimizer that uses random search."""

    def _sample_conv_e_params(self, hyperparams_dict):
        kg_model_config = OrderedDict()
        # Sample params which are dependent on each other
        embedding_dimensions = hyperparams_dict[EMBEDDING_DIM]
        sampled_index = random.choice(range(len(embedding_dimensions)))
        kg_model_config[EMBEDDING_DIM] = embedding_dimensions[sampled_index]
        kg_model_config[CONV_E_HEIGHT] = hyperparams_dict[CONV_E_HEIGHT][sampled_index]
        kg_model_config[CONV_E_WIDTH] = hyperparams_dict[CONV_E_WIDTH][sampled_index]
        kg_model_config[CONV_E_KERNEL_HEIGHT] = hyperparams_dict[CONV_E_KERNEL_HEIGHT][sampled_index]
        kg_model_config[CONV_E_KERNEL_WIDTH] = hyperparams_dict[CONV_E_KERNEL_WIDTH][sampled_index]

        del hyperparams_dict[EMBEDDING_DIM]
        del hyperparams_dict[CONV_E_HEIGHT]
        del hyperparams_dict[CONV_E_WIDTH]
        del hyperparams_dict[CONV_E_KERNEL_HEIGHT]
        del hyperparams_dict[CONV_E_KERNEL_WIDTH]

        kg_model_config.update(self._sample_parameter_value(hyperparams_dict))

        return kg_model_config

    def optimize_hyperparams(self,
                             mapped_train_triples,
                             mapped_test_triples,
                             entity_to_id,
                             rel_to_id,
                             config,
                             device,
                             seed: Optional[int] = None,
                             k_evaluation: int = 10) -> OptimizeResult:
        """"""
        if seed is not None:
            # FIXME np.random is not used
            np.random.seed(seed=seed)

        trained_models: List[Module] = []
        epoch_losses: List[List[float]] = []
        hits_at_k_evaluation: List[float] = []
        entity_to_ids: List = []
        rel_to_ids: List = []
        models_params: List[Dict] = []
        eval_summaries: List = []

        config = config.copy()
        max_iters = config[NUM_OF_HPO_ITERS]

        sample_fct = (
            self._sample_conv_e_params
            if config[KG_EMBEDDING_MODEL_NAME] == CONV_E_NAME else
            self._sample_parameter_value
        )

        for _ in range(max_iters):
            # Sample hyper-params
            kg_embedding_model_config: Dict[int, Any] = sample_fct(config)
            kg_embedding_model_config[NUM_ENTITIES]: int = len(entity_to_id)
            kg_embedding_model_config[NUM_RELATIONS]: int = len(rel_to_id)
            kg_embedding_model_config[SEED]: int = seed

            # Configure defined model
            kg_embedding_model: Module = get_kg_embedding_model(config=kg_embedding_model_config)

            models_params.append(kg_embedding_model_config)
            entity_to_ids.append(entity_to_id)
            rel_to_ids.append(rel_to_id)

            all_entities = np.array(list(entity_to_id.values()))

            trained_model, epoch_loss = train_model(
                kg_embedding_model=kg_embedding_model,
                all_entities=all_entities,
                learning_rate=kg_embedding_model_config[LEARNING_RATE],
                num_epochs=kg_embedding_model_config[NUM_EPOCHS],
                batch_size=kg_embedding_model_config[BATCH_SIZE],
                pos_triples=mapped_train_triples,
                device=device,
                seed=seed,
            )

            # Evaluate trained model
            mean_rank, hits_at_k = compute_metric_results(
                all_entities=all_entities,
                kg_embedding_model=trained_model,
                mapped_train_triples=mapped_train_triples,
                mapped_test_triples=mapped_test_triples,
                device=device,
            )

            # TODO: Define HPO metric
            eval_summary = OrderedDict()
            eval_summary[MEAN_RANK]: float = mean_rank
            eval_summary[HITS_AT_K]: Dict[int, float] = hits_at_k
            eval_summaries.append(eval_summary)

            trained_models.append(trained_model)
            epoch_losses.append(epoch_loss)

            hits_at_k_evaluation.append(hits_at_k[k_evaluation])

        index_of_max = int(np.argmax(a=hits_at_k_evaluation))

        return (
            trained_models[index_of_max],
            epoch_losses[index_of_max],
            entity_to_ids[index_of_max],
            rel_to_ids[index_of_max],
            eval_summaries[index_of_max],
            models_params[index_of_max],
        )

    @classmethod
    def run(cls,
            mapped_train_triples,
            mapped_test_triples,
            entity_to_id,
            rel_to_id,
            config,
            device,
            seed) -> OptimizeResult:
        return cls().optimize_hyperparams(
            mapped_train_triples=mapped_train_triples,
            mapped_test_triples=mapped_test_triples,
            entity_to_id=entity_to_id,
            rel_to_id=rel_to_id,
            config=config,
            device=device,
            seed=seed,
        )




import torch

if __name__ == '__main__':
    heads = torch.tensor([1,3,5,1])
    tails = torch.tensor([2,4,4])

    ent = torch.cat([heads,tails],)
    ent_unique = torch.unique(ent)

    print(ent)
    print(ent_unique)


# -*- coding: utf-8 -*-

import logging

import numpy as np

log = logging.getLogger(__name__)


def create_mapped_triples(triples, entity_to_id=None, rel_to_id=None):
    """

    :param path_to_kg:
    :return:
    """
    if entity_to_id is None or rel_to_id is None:
        entity_to_id, rel_to_id = create_mappings(triples)

    subject_column = np.vectorize(entity_to_id.get)(triples[:, 0:1])
    relation_column = np.vectorize(rel_to_id.get)(triples[:, 1:2])
    object_column = np.vectorize(entity_to_id.get)(triples[:, 2:3])
    triples_of_ids = np.concatenate([subject_column, relation_column, object_column], axis=1)

    triples_of_ids = np.array(triples_of_ids, dtype=np.long)
    # Note: Unique changes the order
    triples_of_ids = np.unique(ar=triples_of_ids, axis=0)

    return triples_of_ids, entity_to_id, rel_to_id


def create_mappings(triples):
    entities = np.unique(np.ndarray.flatten(np.concatenate([triples[:, 0:1], triples[:, 2:3]])))
    relations = np.unique(np.ndarray.flatten(triples[:, 1:2]).tolist())
    entity_to_id = {value: key for key, value in enumerate(entities)}
    rel_to_id = {value: key for key, value in enumerate(relations)}

    return entity_to_id, rel_to_id




import torch
import torch.nn as nn
import numpy as np
import torch.optim as optim

class TransE(nn.Module):
    def __init__(self):
        super(TransE, self).__init__()
        self.entities_embs = torch.nn.Embedding(3, 2)
        pre_trained_entities = np.array([[1., 1.], [2., 2.], [40., -500.]])
        self.entities_embs.weight.data.copy_(torch.from_numpy(pre_trained_entities))

        self.relation_embs = torch.nn.Embedding(1, 2)
        pre_trained_relation = np.array([[3., 3.]])
        self.relation_embs.weight.data.copy_(torch.from_numpy(pre_trained_relation))

        print("############## Initialised Embeddigns#############")
        print(self.entities_embs.weight.data)
        print()
        print(self.relation_embs.weight.data)

        self.criterion = nn.MarginRankingLoss(margin=3., size_average=True)

    def compute_score(self, h_embs, r_embs, t_embs):
        """

        :param h_embs:
        :param r_embs:
        :param t_embs:
        :return:
        """

        # Add the vector element wise
        sum_res = h_embs + r_embs - t_embs

        # Square root
        square_res = torch.mul(sum_res, sum_res)
        reduced_sum_res = torch.sum(square_res, 1)

        # Take the square root element wise
        sqrt_res = torch.sqrt(reduced_sum_res)
        # The scores are the negative distane
        distances = sqrt_res

        n = torch.norm(sum_res).view(size=(-1,))
        print("score norm shape: ", n.shape)
        print("distances shape: ", distances.shape)
        print("score norm: ", n)
        print("distances shape: ", distances)


        return distances

    def compute_loss(self, pos_scores, neg_scores):
        """

        :param pos_scores:
        :param neg_scores:
        :return:
        """

        # y == -1 indicates that second input to criterion should get a larger loss
        # y = torch.Tensor([-1]).cuda()
        # NOTE: y = 1 is important
        y = np.repeat([-1],repeats=pos_scores.shape[0])
        # y = torch.tensor([-1], dtype=torch.float)
        y = torch.tensor(y, dtype=torch.float)

        # Scores for the psotive and negative triples
        pos_scores = torch.tensor(pos_scores, dtype=torch.float)
        neg_scores = torch.tensor(neg_scores, dtype=torch.float)

        loss = self.criterion(pos_scores, neg_scores, y)

        return loss


if __name__ == '__main__':

    trans_e = TransE()
    print('##################################################')

    print()
    print('###########Compute Score###########')
    pos_heads = torch.tensor([[2]],dtype=torch.long)
    pos_tails = torch.tensor([[0]], dtype=torch.long)
    rels = torch.tensor([[0]], dtype=torch.long)

    neg_heads = torch.tensor([[2]], dtype=torch.long)
    neg_tails = torch.tensor([[2]], dtype=torch.long)

    pos_h_embs = trans_e.entities_embs(pos_heads).view(-1,2)
    pos_t_embs = trans_e.entities_embs(pos_tails).view(-1,2)
    neg_h_embs = trans_e.entities_embs(neg_heads).view(-1, 2)
    neg_t_embs = trans_e.entities_embs(neg_tails).view(-1, 2)
    rel_embs = trans_e.relation_embs(rels).view(-1,2)

    pos_score = trans_e.compute_score(pos_h_embs, rel_embs, pos_t_embs)
    neg_score = 1 * trans_e.compute_score(neg_h_embs, rel_embs, neg_t_embs)

    print("pos_score: ", pos_score)
    print("neg_score: ", neg_score)

    print("###########Apply SGD#########")
    optimizer = optim.SGD(trans_e.parameters(), lr=0.1)

    optimizer.zero_grad()

    loss = trans_e.compute_loss(pos_score,neg_score)
    # loss = loss.copy(1.000590)
    # loss = torch.tensor([1.000590],dtype=torch.float)

    print("loss: ", loss)

    loss.backward()

    parameters = filter(lambda p: p.requires_grad, trans_e.parameters())
    print("params: ",parameters)
    for p in parameters:
        print(p)
    print(trans_e.parameters())
    exit(0)
    for p in trans_e.parameters():
        if p.requires_grad:
            print(p.grad)






import matplotlib
matplotlib.matplotlib_fname()
import matplotlib.pyplot as plt
import numpy as np

if __name__ == '__main__':
    losses = [1, 0.5, 0.25, 0.3, 1]
    epochs = np.arange(start=1, stop=len(losses) + 1)
    plt.title(r'Loss Per Epoch')
    plt.xlabel('epoch')
    plt.ylabel('loss')

    plt.plot(epochs, losses)
    plt.savefig('/Users/mehdi/PycharmProjects/kg_embeddings_pipeline/data/corpora/out/losses.png')




import numpy as np
import torch

if __name__ == '__main__':
    entities = np.array([0,1,2,3],dtype=np.int)
    relations = [0,1]
    test_triples = np.array([[0,0,1], [0,1,2], [1,0,1]],dtype=np.int)

    for row_nmbr, row in enumerate(test_triples):
        candidate_entities_subject_based = np.delete(arr=entities, obj=row[0:1])
        candidate_entities_subject_based = np.reshape(candidate_entities_subject_based, newshape=(-1, 1))
        candidate_entities_object_based = np.delete(arr=entities, obj=row[2:3])
        candidate_entities_object_based = np.reshape(candidate_entities_object_based, newshape=(-1, 1))

        print("candidate_entities_subject_based.T : ", candidate_entities_subject_based.T)
        print("candidate_entities_object_based.T: ", candidate_entities_object_based.T)
        print()

        tuple_subject_based = np.reshape(a=test_triples[row_nmbr, 1:3], newshape=(1, 2))
        tuple_object_based = np.reshape(a=test_triples[row_nmbr, 0:2], newshape=(1, 2))

        print("tuple_subject_based: ", tuple_subject_based)
        print("tuple_object_based: ", tuple_object_based)
        print()

        tuples_subject_based = np.repeat(a=tuple_subject_based, repeats=candidate_entities_subject_based.shape[0],
                                         axis=0)
        tuples_object_based = np.repeat(a=tuple_object_based, repeats=candidate_entities_object_based.shape[0], axis=0)

        print("tuples_subject_based: ", tuples_subject_based)
        print("tuples_object_based: ", tuples_object_based)
        print()

        corrupted_subject_based = np.concatenate([candidate_entities_subject_based, tuples_subject_based], axis=1)
        corrupted_subject_based = torch.tensor(corrupted_subject_based, dtype=torch.long)

        corrupted_object_based = np.concatenate([tuples_object_based, candidate_entities_object_based], axis=1)
        corrupted_object_based = torch.tensor(corrupted_object_based, dtype=torch.long)

        print("corrupted_subject_based: ", corrupted_subject_based)
        print("corrupted_object_based: ", corrupted_object_based)
        print()

        pos_triple = np.array(row)
        pos_triple = np.expand_dims(a=pos_triple, axis=0)
        pos_triple = torch.tensor(pos_triple, dtype=torch.long)
        score_of_positive = 3.2

        print("pos_triple: ", pos_triple)

        scores_of_corrupted_subjects = [2.5, 0.5, 10.]
        scores_object_based = [2.5, 0.5, 10.]

        scores_subject_based = np.append(arr=scores_of_corrupted_subjects, values=score_of_positive)
        indice_of_pos_subject_based = scores_subject_based.size - 1

        print("scores_subject_based: ", scores_subject_based)
        print("indice_of_pos_subject_based: ", indice_of_pos_subject_based)
        print()

        _, sorted_score_indices_subject_based = torch.sort(torch.tensor(scores_subject_based, dtype=torch.float),
                                                           descending=False)
        sorted_score_indices_subject_based = sorted_score_indices_subject_based.cpu().numpy()

        print("sorted_score_indices_subject_based: ", sorted_score_indices_subject_based)
        print()

        rank_of_positive_subject_based = np.where(sorted_score_indices_subject_based == indice_of_pos_subject_based)[0][
            0]

        print("rank_of_positive_subject_based: ", rank_of_positive_subject_based)

        exit(0)


# -*- coding: utf-8 -*-

"""PyKEEN's command line interface."""

from pykeen.constants import (
    CONV_E_NAME, DISTMULT_NAME, ERMLP_NAME, RESCAL_NAME, SE_NAME, TRANS_D_NAME, TRANS_E_NAME, TRANS_H_NAME,
    TRANS_R_NAME, UM_NAME,
)
from pykeen.utilities.cli_utils import (
    configure_conv_e_hpo_pipeline, configure_conv_e_training_pipeline,
    configure_distmult_hpo_pipeline, configure_distmult_training_pipeline, configure_ermlp_hpo_pipeline,
    configure_ermlp_training_pipeline, configure_rescal_hpo_pipeline, configure_rescal_training_pipeline,
    configure_se_hpo_pipeline, configure_se_training_pipeline, configure_trans_d_hpo_pipeline,
    configure_trans_d_training_pipeline, configure_trans_e_hpo_pipeline, configure_trans_e_training_pipeline,
    configure_trans_h_hpo_pipeline, configure_trans_h_training_pipeline, configure_trans_r_hpo_pipeline,
    configure_trans_r_training_pipeline, configure_um_hpo_pipeline, configure_um_training_pipeline,
)

__all__ = [
    'MODEL_TRAINING_CONFIG_FUNCS',
    'MODEL_HPO_CONFIG_FUNCS',
]

MODEL_TRAINING_CONFIG_FUNCS = {
    TRANS_E_NAME: configure_trans_e_training_pipeline,
    TRANS_H_NAME: configure_trans_h_training_pipeline,
    TRANS_R_NAME: configure_trans_r_training_pipeline,
    TRANS_D_NAME: configure_trans_d_training_pipeline,
    SE_NAME: configure_se_training_pipeline,
    UM_NAME: configure_um_training_pipeline,
    DISTMULT_NAME: configure_distmult_training_pipeline,
    ERMLP_NAME: configure_ermlp_training_pipeline,
    RESCAL_NAME: configure_rescal_training_pipeline,
    CONV_E_NAME: configure_conv_e_training_pipeline
}

MODEL_HPO_CONFIG_FUNCS = {
    TRANS_E_NAME: configure_trans_e_hpo_pipeline,
    TRANS_H_NAME: configure_trans_h_hpo_pipeline,
    TRANS_R_NAME: configure_trans_r_hpo_pipeline,
    TRANS_D_NAME: configure_trans_d_hpo_pipeline,
    SE_NAME: configure_se_hpo_pipeline,
    UM_NAME: configure_um_hpo_pipeline,
    DISTMULT_NAME: configure_distmult_hpo_pipeline,
    ERMLP_NAME: configure_ermlp_hpo_pipeline,
    RESCAL_NAME: configure_rescal_hpo_pipeline,
    CONV_E_NAME: configure_conv_e_hpo_pipeline
}



# -*- coding: utf-8 -*-

"""Script for starting the pipeline and saving the results."""

import json
import os
import pickle
import time
from dataclasses import dataclass
from typing import Dict, Mapping, Optional

import torch

from pykeen.constants import (
    ENTITY_TO_EMBEDDING, ENTITY_TO_ID, EVAL_SUMMARY, LOSSES, OUTPUT_DIREC, RELATION_TO_EMBEDDING, RELATION_TO_ID,
    TRAINED_MODEL,
    FINAL_CONFIGURATION)
from pykeen.utilities.pipeline import Pipeline


@dataclass
class Results:
    """Results from PyKEEN."""

    config: Mapping
    pipeline: Pipeline
    results: Mapping


def run(config: Dict,
        output_directory: Optional[str] = None) -> Results:
    """Run PyKEEN using a given configuration."""
    if output_directory is None:
        output_directory = os.path.join(config[OUTPUT_DIREC], time.strftime("%Y-%m-%d_%H:%M:%S"))
    os.makedirs(output_directory, exist_ok=True)

    pipeline = Pipeline(config=config)
    pipeline_results = pipeline.run()

    with open(os.path.join(output_directory, 'configuration.json'), 'w') as file:
        # In HPO model inital configuration is different from final configurations, thats why we differentiate
        json.dump(pipeline_results[FINAL_CONFIGURATION], file, indent=2)

    with open(os.path.join(output_directory, 'entities_to_embeddings.pkl'), 'wb') as file:
        pickle.dump(pipeline_results[ENTITY_TO_EMBEDDING], file, protocol=pickle.HIGHEST_PROTOCOL)

    with open(os.path.join(output_directory, 'entities_to_embeddings.json'), 'w') as file:
        json.dump(
            {
                key: list(map(float, array))
                for key, array in pipeline_results[ENTITY_TO_EMBEDDING].items()
            },
            file,
            indent=2,
            sort_keys=True
        )

    with open(os.path.join(output_directory, 'relations_to_embeddings.pkl'), 'wb') as file:
        pickle.dump(pipeline_results[RELATION_TO_EMBEDDING], file, protocol=pickle.HIGHEST_PROTOCOL)

    with open(os.path.join(output_directory, 'relations_to_embeddings.json'), 'w') as file:
        json.dump(
            {
                key: list(map(float, array))
                for key, array in pipeline_results[RELATION_TO_EMBEDDING].items()
            },
            file,
            indent=2,
            sort_keys=True,
        )

    with open(os.path.join(output_directory, 'entity_to_id.json'), 'w') as file:
        json.dump(pipeline_results[ENTITY_TO_ID], file, indent=2, sort_keys=True)

    with open(os.path.join(output_directory, 'relation_to_id.json'), 'w') as file:
        json.dump(pipeline_results[RELATION_TO_ID], file, indent=2, sort_keys=True)

    with open(os.path.join(output_directory, 'losses.json'), 'w') as file:
        json.dump(pipeline_results[LOSSES], file, indent=2, sort_keys=True)

    eval_summary = pipeline_results.get(EVAL_SUMMARY)
    if eval_summary is not None:
        with open(os.path.join(output_directory, 'evaluation_summary.json'), 'w') as file:
            json.dump(eval_summary, file, indent=2)

    # Save trained model
    torch.save(
        pipeline_results[TRAINED_MODEL].state_dict(),
        os.path.join(output_directory, 'trained_model.pkl'),
    )

    return Results(
        config=config,
        pipeline=pipeline,
        results=pipeline_results,
    )




import numpy as np
import torch

def split_list_in_batches(input_list, batch_size):
    return [input_list[i:i + batch_size] for i in range(0, len(input_list), batch_size)]

if __name__ == '__main__':
    batch_size = 2
    all_entities = np.array([0, 1, 2, 3], dtype=np.int)
    relations = [0, 1]
    pos_triples = np.array([[0, 0, 1], [0, 1, 2], [1, 0, 1]], dtype=np.int)

    num_pos_triples = pos_triples.shape[0]
    num_entities = all_entities.shape[0]


    for epoch in range(1):
        print("Before shuffling pos_triples: ", pos_triples)
        print()
        np.random.seed(seed=2)
        indices = np.arange(num_pos_triples)
        np.random.shuffle(indices)
        pos_triples = pos_triples[indices]
        print("After shuffling pos_triples: ", pos_triples)
        print()

        pos_batches = split_list_in_batches(input_list=pos_triples, batch_size=batch_size)

        print('pos_batches: ', pos_batches)
        print()

        for i in range(len(pos_batches)):
            pos_batch = pos_batches[i]

            print('pos_batch: ', pos_batch)

            current_batch_size = len(pos_batch)
            print('current_batch_size: ', current_batch_size)
            batch_subjs = pos_batch[:, 0:1]
            batch_relations = pos_batch[:, 1:2]
            batch_objs = pos_batch[:, 2:3]

            print("batch_subjs: ", batch_subjs)
            print('batch_relations: ', batch_relations)
            print('batch_objs: ', batch_objs)

            print()

            num_subj_corrupt = len(pos_batch) // 2
            num_obj_corrupt = len(pos_batch) - num_subj_corrupt
            pos_batch = torch.tensor(pos_batch, dtype=torch.long)

            print('num_subj_corrupt: ', num_subj_corrupt)
            print('num_obj_corrupt: ', num_obj_corrupt)
            print()

            corrupted_subj_indices = np.random.choice(np.arange(0, num_entities), size=num_subj_corrupt)
            print('corrupted_subj_indices: ', corrupted_subj_indices)

            corrupted_subjects = np.reshape(all_entities[corrupted_subj_indices], newshape=(-1, 1))

            print('corrupted_subjects: ', corrupted_subjects)

            subject_based_corrupted_triples = np.concatenate(
                [corrupted_subjects, batch_relations[:num_subj_corrupt], batch_objs[:num_subj_corrupt]], axis=1)

            print('subject_based_corrupted_triples: ', subject_based_corrupted_triples)

            corrupted_obj_indices = np.random.choice(np.arange(0, num_entities), size=num_obj_corrupt)
            print('corrupted_obj_indices: ', corrupted_obj_indices)

            corrupted_objects = np.reshape(all_entities[corrupted_obj_indices], newshape=(-1, 1))
            print('corrupted_objects: ', corrupted_objects)

            object_based_corrupted_triples = np.concatenate(
                [batch_subjs[num_subj_corrupt:], batch_relations[num_subj_corrupt:], corrupted_objects], axis=1)

            print('object_based_corrupted_triples: ', object_based_corrupted_triples)

            neg_batch = np.concatenate([subject_based_corrupted_triples, object_based_corrupted_triples], axis=0)

            neg_batch = torch.tensor(neg_batch, dtype=torch.long)

            print('neg_batch: ', neg_batch)

            print()

        print('###########')